from libc.stdint cimport int8_t, int16_t, int32_t, int64_t, intptr_t
from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t, uintptr_t
cdef extern from *:
  ctypedef bint bool
  ctypedef struct va_list

cdef extern from *:

  cdef enum:
    IV,
  ctypedef ptrdiff_t IE;

  cdef enum:
    UV,
  ctypedef size_t UE;

  ctypedef size_t Usize;

  ctypedef ptrdiff_t Isize;

  void root(Usize, Isize, UE, IE);
